from __future__ import annotations

import warnings
from typing import TYPE_CHECKING

from safeds._utils import _figure_to_image
from safeds._validation import _check_columns_exist
from safeds.exceptions import NonNumericColumnError

if TYPE_CHECKING:
    from safeds.data.image.containers import Image
    from safeds.data.tabular.containers import Table


class TablePlotter:
    """
    A class that contains plotting methods for a table.

    Parameters
    ----------
    table:
        The table to plot.

    Examples
    --------
    >>> from safeds.data.tabular.containers import Table
    >>> table = Table({"test": [1, 2, 3]})
    >>> plotter = table.plot
    """

    def __init__(self, table: Table):
        self._table: Table = table

    def box_plots(self) -> Image:
        """
        Plot a boxplot for every numerical column.

        Returns
        -------
        plot:
            The plot as an image.

        Raises
        ------
        NonNumericColumnError
            If the table contains only non-numerical columns.

        Examples
        --------
        >>> from safeds.data.tabular.containers import Table
        >>> table = Table({"a":[1, 2], "b": [3, 42]})
        >>> image = table.plot.box_plots()
        """
        # TODO: implement using matplotlib and polars
        import matplotlib.pyplot as plt
        import seaborn as sns

        numerical_table = self._table.remove_non_numeric_columns()
        if numerical_table.number_of_columns == 0:
            raise NonNumericColumnError("This table contains only non-numerical columns.")
        col_wrap = min(numerical_table.number_of_columns, 3)

        data = numerical_table._lazy_frame.melt(value_vars=numerical_table.column_names).collect()
        grid = sns.FacetGrid(data, col="variable", col_wrap=col_wrap, sharex=False, sharey=False)
        with warnings.catch_warnings():
            warnings.filterwarnings(
                "ignore",
                message="Using the boxplot function without specifying `order` is likely to produce an incorrect plot.",
            )
            grid.map(sns.boxplot, "variable", "value")
        grid.set_xlabels("")
        grid.set_ylabels("")
        grid.set_titles("{col_name}")
        for axes in grid.axes.flat:
            axes.set_xticks([])
        plt.tight_layout()
        fig = grid.fig

        return _figure_to_image(fig)

    def correlation_heatmap(self) -> Image:
        """
        Plot a correlation heatmap for all numerical columns of this `Table`.

        Returns
        -------
        plot:
            The plot as an image.

        Examples
        --------
        >>> from safeds.data.tabular.containers import Table
        >>> table = Table.from_dict({"temperature": [10, 15, 20, 25, 30], "sales": [54, 74, 90, 206, 210]})
        >>> image = table.plot.correlation_heatmap()
        """
        # TODO: implement using matplotlib and polars
        #  https://stackoverflow.com/questions/33282368/plotting-a-2d-heatmap
        import matplotlib.pyplot as plt
        import seaborn as sns

        only_numerical = self._table.remove_non_numeric_columns()._data_frame.fill_null(0)

        if self._table.number_of_rows == 0:
            warnings.warn(
                "An empty table has been used. A correlation heatmap on an empty table will show nothing.",
                stacklevel=2,
            )

        with warnings.catch_warnings():
            warnings.filterwarnings(
                "ignore",
                message=(
                    "Attempting to set identical low and high (xlims|ylims) makes transformation singular;"
                    " automatically expanding."
                ),
            )
            fig = plt.figure()
            sns.heatmap(
                data=only_numerical.corr().to_numpy(),
                vmin=-1,
                vmax=1,
                xticklabels=only_numerical.columns,
                yticklabels=only_numerical.columns,
                cmap="vlag",
            )
            plt.tight_layout()

        return _figure_to_image(fig)

    def histograms(self, *, maximum_number_of_bins: int = 10) -> Image:
        """
        Plot a histogram for every column.

        Parameters
        ----------
        maximum_number_of_bins:
            The maximum number of bins to use in the histogram. Default is 10.

        Returns
        -------
        plot:
            The plot as an image.

        Examples
        --------
        >>> from safeds.data.tabular.containers import Table
        >>> table = Table({"a": [2, 3, 5, 1], "b": [54, 74, 90, 2014]})
        >>> image = table.plot.histograms()
        """
        import matplotlib.pyplot as plt
        import polars as pl

        n_cols = min(3, self._table.number_of_columns)
        n_rows = 1 + (self._table.number_of_columns - 1) // n_cols

        if n_cols == 1 and n_rows == 1:
            fig, axs = plt.subplots(1, 1, tight_layout=True)
            one_col = True
        else:
            fig, axs = plt.subplots(n_rows, n_cols, tight_layout=True, figsize=(n_cols * 3, n_rows * 3))
            one_col = False

        col_names = self._table.column_names
        for col_name, ax in zip(col_names, axs.flatten() if not one_col else [axs], strict=False):
            column = self._table.get_column(col_name)
            distinct_values = column.get_distinct_values()

            ax.set_title(col_name)
            ax.set_xlabel("")
            ax.set_ylabel("")

            if column.is_numeric and len(distinct_values) > maximum_number_of_bins:
                min_val = (column.min() or 0) - 1e-6  # Otherwise the minimum value is not included in the first bin
                max_val = column.max() or 0
                bin_count = min(maximum_number_of_bins, len(distinct_values))
                bins = [
                    *(pl.Series(range(bin_count + 1)) / bin_count * (max_val - min_val) + min_val),
                ]

                bars = [f"{round((bins[i] + bins[i + 1]) / 2, 2)}" for i in range(len(bins) - 1)]
                hist = (
                    column._series.hist(bins=bins)
                    .slice(1, length=maximum_number_of_bins)
                    .get_column("count")
                    .to_numpy()
                )

                ax.bar(bars, hist, edgecolor="black")
                ax.set_xticks(range(len(hist)), bars, rotation=45, horizontalalignment="right")
            else:
                value_counts = (
                    column._series.drop_nulls().value_counts().sort(column.name).slice(0, length=maximum_number_of_bins)
                )
                distinct_values = value_counts.get_column(column.name).cast(pl.String).to_numpy()
                hist = value_counts.get_column("count").to_numpy()
                ax.bar(distinct_values, hist, edgecolor="black")
                ax.set_xticks(range(len(distinct_values)), distinct_values, rotation=45, horizontalalignment="right")

        for i in range(len(col_names), n_rows * n_cols):
            fig.delaxes(axs.flatten()[i])  # Remove empty subplots

        return _figure_to_image(fig)

    def line_plot(self, x_name: str, y_name: str) -> Image:
        """
        Create a line plot for two columns in the table.

        Parameters
        ----------
        x_name:
            The name of the column to be plotted on the x-axis.
        y_name:
            The name of the column to be plotted on the y-axis.

        Returns
        -------
        line_plot:
            The plot as an image.

        Raises
        ------
        ColumnNotFoundError
            If a column does not exist.
        TypeError
            If a column is not numeric.

        Examples
        --------
        >>> from safeds.data.tabular.containers import Table
        >>> table = Table(
        ...     {
        ...         "a": [1, 2, 3, 4, 5],
        ...         "b": [2, 3, 4, 5, 6],
        ...     }
        ... )
        >>> image = table.plot.line_plot("a", "b")
        """
        _check_columns_exist(self._table, [x_name, y_name])

        # TODO: pass list of columns names + extract validation
        if not self._table.get_column(x_name).is_numeric:
            raise NonNumericColumnError(x_name)
        if not self._table.get_column(y_name).is_numeric:
            raise NonNumericColumnError(y_name)

        import matplotlib.pyplot as plt
        import polars as pl

        grouped = (
            self._table._lazy_frame.group_by(x_name, maintain_order=True)
            .agg(
                mean=pl.mean(y_name),
                count=pl.count(y_name),
                standard_deviation=pl.std(y_name, ddof=0),
            )
            .collect()
        )

        x = grouped.get_column(x_name)
        y = grouped.get_column("mean")
        confidence_interval = 1.96 * grouped.get_column("standard_deviation") / grouped.get_column("count").sqrt()

        fig, ax = plt.subplots()
        ax.plot(
            x,
            y,
        )
        ax.fill_between(
            x,
            y - confidence_interval,
            y + confidence_interval,
            color="lightblue",
            alpha=0.15,
        )
        ax.set(
            xlabel=x_name,
            ylabel=y_name,
        )
        ax.set_xticks(ax.get_xticks())
        ax.set_xticklabels(
            ax.get_xticklabels(),
            rotation=45,
            horizontalalignment="right",
        )  # rotate the labels of the x Axis to prevent the chance of overlapping of the labels
        fig.tight_layout()

        return _figure_to_image(fig)

    def scatter_plot(self, x_name: str, y_name: str) -> Image:
        """
        Create a scatter plot for two columns in the table.

        Parameters
        ----------
        x_name:
            The name of the column to be plotted on the x-axis.
        y_name:
            The name of the column to be plotted on the y-axis.

        Returns
        -------
        scatter_plot:
            The plot as an image.

        Raises
        ------
        ColumnNotFoundError
            If a column does not exist.
        TypeError
            If a column is not numeric.

        Examples
        --------
        >>> from safeds.data.tabular.containers import Table
        >>> table = Table(
        ...     {
        ...         "a": [1, 2, 3, 4, 5],
        ...         "b": [2, 3, 4, 5, 6],
        ...     }
        ... )
        >>> image = table.plot.scatter_plot("a", "b")
        """
        _check_columns_exist(self._table, [x_name, y_name])

        # TODO: pass list of columns names + extract validation
        if not self._table.get_column(x_name).is_numeric:
            raise NonNumericColumnError(x_name)
        if not self._table.get_column(y_name).is_numeric:
            raise NonNumericColumnError(y_name)

        import matplotlib.pyplot as plt

        fig, ax = plt.subplots()
        ax.scatter(
            x=self._table.get_column(x_name)._series,
            y=self._table.get_column(y_name)._series,
        )
        ax.set(
            xlabel=x_name,
            ylabel=y_name,
        )
        ax.set_xticks(ax.get_xticks())
        ax.set_xticklabels(
            ax.get_xticklabels(),
            rotation=45,
            horizontalalignment="right",
        )  # rotate the labels of the x Axis to prevent the chance of overlapping of the labels
        fig.tight_layout()

        return _figure_to_image(fig)

    # TODO: equivalent to Column.plot_compare_columns that takes a list of column names (index_plot)?



from __future__ import annotations

from typing import TYPE_CHECKING

from safeds._utils import _get_random_seed, _structural_hash

from ._classifier import Classifier

if TYPE_CHECKING:
    from sklearn.base import ClassifierMixin


class LogisticClassifier(Classifier):
    """Regularized logistic regression for classification."""

    # ------------------------------------------------------------------------------------------------------------------
    # Dunder methods
    # ------------------------------------------------------------------------------------------------------------------

    def __init__(self) -> None:
        super().__init__()

    def __hash__(self) -> int:
        return _structural_hash(
            super().__hash__(),
        )

    # ------------------------------------------------------------------------------------------------------------------
    # Template methods
    # ------------------------------------------------------------------------------------------------------------------

    def _clone(self) -> LogisticClassifier:
        return LogisticClassifier()

    def _get_sklearn_model(self) -> ClassifierMixin:
        from sklearn.linear_model import LogisticRegression as SklearnLogisticRegression

        return SklearnLogisticRegression(
            random_state=_get_random_seed(),
            n_jobs=-1,
        )



from __future__ import annotations

from typing import TYPE_CHECKING, Any

from safeds.data.labeled.containers import TimeSeriesDataset

from ._input_converter import InputConversion

if TYPE_CHECKING:
    from torch.utils.data import DataLoader


class InputConversionTimeSeries(InputConversion[TimeSeriesDataset, TimeSeriesDataset]):
    """The input conversion for a neural network, defines the input parameters for the neural network."""

    def __init__(
        self,
        window_size: int,
        forecast_horizon: int,
    ) -> None:
        """
        Define the input parameters for the neural network in the input conversion.

        Parameters
        ----------
        window_size:
            The size of the created windows
        forecast_horizon:
            The forecast horizon defines the future lag of the predicted values
        """
        self._window_size = window_size
        self._forecast_horizon = forecast_horizon
        self._first = True
        self._target_name: str = ""
        self._time_name: str = ""
        self._feature_names: list[str] = []

    @property
    def _data_size(self) -> int:
        """
        Gives the size for the input of an internal layer.

        Returns
        -------
        size:
            The size of the input for the neural network

        """
        return (len(self._feature_names) + 1) * self._window_size

    def _data_conversion_fit(
        self,
        input_data: TimeSeriesDataset,
        batch_size: int,
        num_of_classes: int = 1,
    ) -> DataLoader:
        self._num_of_classes = num_of_classes
        return input_data._into_dataloader_with_window(
            self._window_size,
            self._forecast_horizon,
            batch_size,
        )

    def _data_conversion_predict(self, input_data: TimeSeriesDataset, batch_size: int) -> DataLoader:
        return input_data._into_dataloader_with_window_predict(self._window_size, self._forecast_horizon, batch_size)

    def _is_fit_data_valid(self, input_data: TimeSeriesDataset) -> bool:
        if self._first:
            self._time_name = input_data.time.name
            self._feature_names = input_data.features.column_names
            self._target_name = input_data.target.name
            self._first = False
        return (
            (sorted(input_data.features.column_names)).__eq__(sorted(self._feature_names))
            and input_data.target.name == self._target_name
            and input_data.time.name == self._time_name
        )

    def _is_predict_data_valid(self, input_data: TimeSeriesDataset) -> bool:
        return self._is_fit_data_valid(input_data)

    def _get_output_configuration(self) -> dict[str, Any]:
        return {"window_size": self._window_size, "forecast_horizon": self._forecast_horizon}



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:
    from torch import nn

    from safeds.data.image.typing import ImageSize


class Layer(ABC):
    @abstractmethod
    def __init__(self) -> None:
        pass  # pragma: no cover

    @abstractmethod
    def _get_internal_layer(self, **kwargs: Any) -> nn.Module:
        pass  # pragma: no cover

    @property
    @abstractmethod
    def input_size(self) -> int | ImageSize:
        pass  # pragma: no cover

    @property
    @abstractmethod
    def output_size(self) -> int | ImageSize:
        pass  # pragma: no cover

    @abstractmethod
    def _set_input_size(self, input_size: int | ImageSize) -> None:
        pass  # pragma: no cover

    @abstractmethod
    def __hash__(self) -> int:
        pass  # pragma: no cover

    @abstractmethod
    def __eq__(self, other: object) -> bool:
        pass  # pragma: no cover

    @abstractmethod
    def __sizeof__(self) -> int:
        pass  # pragma: no cover



from __future__ import annotations

import math
import sys
from typing import TYPE_CHECKING, Any, Literal

from safeds._config import _init_default_device
from safeds._utils import _structural_hash
from safeds.data.image.typing import ImageSize

from ._layer import Layer

if TYPE_CHECKING:
    from torch import Tensor, nn


def _create_internal_model(strategy: Literal["max", "avg"], kernel_size: int, padding: int, stride: int) -> nn.Module:
    from torch import nn

    _init_default_device()

    class _InternalLayer(nn.Module):
        def __init__(self, strategy: Literal["max", "avg"], kernel_size: int, padding: int, stride: int):
            super().__init__()
            match strategy:
                case "max":
                    self._layer = nn.MaxPool2d(kernel_size=kernel_size, padding=padding, stride=stride)
                case "avg":
                    self._layer = nn.AvgPool2d(kernel_size=kernel_size, padding=padding, stride=stride)

        def forward(self, x: Tensor) -> Tensor:
            return self._layer(x)

    return _InternalLayer(strategy, kernel_size, padding, stride)


class _Pooling2DLayer(Layer):
    def __init__(self, strategy: Literal["max", "avg"], kernel_size: int, *, stride: int = -1, padding: int = 0):
        """
        Create a Pooling 2D Layer.

        Parameters
        ----------
        strategy:
            the strategy of the pooling
        kernel_size:
            the size of the kernel
        stride:
            the stride of the pooling
        padding:
            the padding of the pooling
        """
        self._strategy = strategy
        self._kernel_size = kernel_size
        self._stride = stride if stride != -1 else kernel_size
        self._padding = padding
        self._input_size: ImageSize | None = None
        self._output_size: ImageSize | None = None

    def _get_internal_layer(self, **kwargs: Any) -> nn.Module:  # noqa: ARG002
        return _create_internal_model(self._strategy, self._kernel_size, self._padding, self._stride)

    @property
    def input_size(self) -> ImageSize:
        """
        Get the input_size of this layer.

        Returns
        -------
        result:
            The amount of values being passed into this layer.

        Raises
        ------
        ValueError
            If the input_size is not yet set
        """
        if self._input_size is None:
            raise ValueError("The input_size is not yet set.")
        return self._input_size

    @property
    def output_size(self) -> ImageSize:
        """
        Get the output_size of this layer.

        Returns
        -------
        result:
            The Number of Neurons in this layer.

        Raises
        ------
        ValueError
            If the input_size is not yet set
        """
        if self._input_size is None:
            raise ValueError(
                "The input_size is not yet set. The layer cannot compute the output_size if the input_size is not set.",
            )
        if self._output_size is None:
            new_width = math.ceil(
                (self.input_size.width + self._padding * 2 - self._kernel_size + 1) / (1.0 * self._stride),
            )
            new_height = math.ceil(
                (self.input_size.height + self._padding * 2 - self._kernel_size + 1) / (1.0 * self._stride),
            )
            self._output_size = ImageSize(new_width, new_height, self._input_size.channel, _ignore_invalid_channel=True)
        return self._output_size

    def _set_input_size(self, input_size: int | ImageSize) -> None:
        if isinstance(input_size, int):
            raise TypeError("The input_size of a pooling layer has to be of type ImageSize.")
        self._input_size = input_size
        self._output_size = None

    def __hash__(self) -> int:
        """
        Return a deterministic hash value for this pooling 2d layer.

        Returns
        -------
        hash:
            the hash value
        """
        return _structural_hash(
            self._strategy,
            self._kernel_size,
            self._stride,
            self._padding,
            self._input_size,
            self._output_size,
        )

    def __eq__(self, other: object) -> bool:
        """
        Compare two pooling 2d layer.

        Parameters
        ----------
        other:
            The pooling 2d layer to compare to.

        Returns
        -------
        equals:
            Whether the two pooling 2d layer are the same.
        """
        if not isinstance(other, type(self)):
            return NotImplemented
        return (self is other) or (
            self._input_size == other._input_size
            and self._output_size == other._output_size
            and self._strategy == other._strategy
            and self._kernel_size == other._kernel_size
            and self._stride == other._stride
            and self._padding == other._padding
        )

    def __sizeof__(self) -> int:
        """
        Return the complete size of this object.

        Returns
        -------
        size:
            Size of this object in bytes.
        """
        return (
            sys.getsizeof(self._input_size)
            + sys.getsizeof(self._output_size)
            + sys.getsizeof(self._strategy)
            + sys.getsizeof(self._kernel_size)
            + sys.getsizeof(self._stride)
            + sys.getsizeof(self._padding)
        )


class MaxPooling2DLayer(_Pooling2DLayer):
    def __init__(self, kernel_size: int, *, stride: int = -1, padding: int = 0) -> None:
        """
        Create a maximum Pooling 2D Layer.

        Parameters
        ----------
        kernel_size:
            the size of the kernel
        stride:
            the stride of the pooling
        padding:
            the padding of the pooling
        """
        super().__init__("max", kernel_size, stride=stride, padding=padding)


class AveragePooling2DLayer(_Pooling2DLayer):
    def __init__(self, kernel_size: int, *, stride: int = -1, padding: int = 0) -> None:
        """
        Create a average Pooling 2D Layer.

        Parameters
        ----------
        kernel_size:
            the size of the kernel
        stride:
            the stride of the pooling
        padding:
            the padding of the pooling
        """
        super().__init__("avg", kernel_size, stride=stride, padding=padding)



from __future__ import annotations

from typing import TYPE_CHECKING

from safeds._validation import _check_columns_exist
from safeds._validation._check_columns_are_numeric import _check_columns_are_numeric
from safeds.data.tabular.containers import Table
from safeds.exceptions import TransformerNotFittedError

from ._invertible_table_transformer import InvertibleTableTransformer

if TYPE_CHECKING:
    import polars as pl


class StandardScaler(InvertibleTableTransformer):
    """The StandardScaler transforms column values to a range by removing the mean and scaling to unit variance."""

    # ------------------------------------------------------------------------------------------------------------------
    # Dunder methods
    # ------------------------------------------------------------------------------------------------------------------

    def __init__(self) -> None:
        super().__init__()

        # Internal state
        self._data_mean: pl.DataFrame | None = None
        self._data_standard_deviation: pl.DataFrame | None = None

    def __hash__(self) -> int:
        # Leave out the internal state for faster hashing
        return super().__hash__()

    # ------------------------------------------------------------------------------------------------------------------
    # Learning and transformation
    # ------------------------------------------------------------------------------------------------------------------

    def fit(self, table: Table, column_names: list[str] | None) -> StandardScaler:
        """
        Learn a transformation for a set of columns in a table.

        This transformer is not modified.

        Parameters
        ----------
        table:
            The table used to fit the transformer.
        column_names:
            The list of columns from the table used to fit the transformer. If `None`, all columns are used.

        Returns
        -------
        fitted_transformer:
            The fitted transformer.

        Raises
        ------
        ColumnNotFoundError
            If column_names contain a column name that is missing in the table.
        ColumnTypeError
            If at least one of the specified columns in the table contains non-numerical data.
        ValueError
            If the table contains 0 rows.
        """
        if column_names is None:
            column_names = [name for name in table.column_names if table.get_column_type(name).is_numeric]
        else:
            _check_columns_exist(table, column_names)
            _check_columns_are_numeric(table, column_names, operation="fit a StandardScaler")

        if table.number_of_rows == 0:
            raise ValueError("The StandardScaler cannot be fitted because the table contains 0 rows")

        # Learn the transformation (ddof=0 is used to match the behavior of scikit-learn)
        _data_mean = table._lazy_frame.select(column_names).mean().collect()
        _data_standard_deviation = table._lazy_frame.select(column_names).std(ddof=0).collect()

        # Create a copy with the learned transformation
        result = StandardScaler()
        result._column_names = column_names
        result._data_mean = _data_mean
        result._data_standard_deviation = _data_standard_deviation

        return result

    def transform(self, table: Table) -> Table:
        """
        Apply the learned transformation to a table.

        The table is not modified.

        Parameters
        ----------
        table:
            The table to which the learned transformation is applied.

        Returns
        -------
        transformed_table:
            The transformed table.

        Raises
        ------
        TransformerNotFittedError
            If the transformer has not been fitted yet.
        ColumnNotFoundError
            If the input table does not contain all columns used to fit the transformer.
        ColumnTypeError
            If at least one of the columns in the input table that is used to fit contains non-numerical data.
        """
        import polars as pl

        # Used in favor of is_fitted, so the type checker is happy
        if self._column_names is None or self._data_mean is None or self._data_standard_deviation is None:
            raise TransformerNotFittedError

        _check_columns_exist(table, self._column_names)
        _check_columns_are_numeric(table, self._column_names, operation="transform with a StandardScaler")

        columns = [
            (pl.col(name) - self._data_mean.get_column(name)) / self._data_standard_deviation.get_column(name)
            for name in self._column_names
        ]

        return Table._from_polars_lazy_frame(
            table._lazy_frame.with_columns(columns),
        )

    def inverse_transform(self, transformed_table: Table) -> Table:
        """
        Undo the learned transformation.

        The table is not modified.

        Parameters
        ----------
        transformed_table:
            The table to be transformed back to the original version.

        Returns
        -------
        original_table:
            The original table.

        Raises
        ------
        TransformerNotFittedError
            If the transformer has not been fitted yet.
        ColumnNotFoundError
            If the input table does not contain all columns used to fit the transformer.
        ColumnTypeError
            If the transformed columns of the input table contain non-numerical data.
        """
        import polars as pl

        # Used in favor of is_fitted, so the type checker is happy
        if self._column_names is None or self._data_mean is None or self._data_standard_deviation is None:
            raise TransformerNotFittedError

        _check_columns_exist(transformed_table, self._column_names)
        _check_columns_are_numeric(
            transformed_table,
            self._column_names,
            operation="inverse-transform with a StandardScaler",
        )

        columns = [
            pl.col(name) * self._data_standard_deviation.get_column(name) + self._data_mean.get_column(name)
            for name in self._column_names
        ]

        return Table._from_polars_lazy_frame(
            transformed_table._lazy_frame.with_columns(columns),
        )
